import { ESLintUtils } from "@typescript-eslint/utils";
import ts from "typescript";
import { createRule } from "../utils/createRule.js";
import { isDiagnosticWithStart } from "../utils/diagnostics.js";
import { loc } from "../utils/loc.js";
import { getTypeSnapshot, updateTypeSnapshot } from "../utils/snapshot.js";
const messages = {
  ExpectedErrorNotFound: "Expected an error on this line, but found none.",
  FileIsNotIncludedInTsconfig: 'Expected to find a file "{{ fileName }}" present.',
  Multiple$ExpectTypeAssertions: "This line has 2 or more $ExpectType assertions.",
  OrphanAssertion: "Can not match a node to this assertion.",
  SyntaxError: "Syntax Error: {{ message }}",
  TypeSnapshotDoNotMatch: "Expected type from Snapshot to be: {{ expected }}, got: {{ actual }}",
  TypeSnapshotNotFound: "Type Snapshot not found. Please consider running ESLint in FIX mode: eslint --fix",
  TypesDoNotMatch: "Expected type to be: {{ expected }}, got: {{ actual }}"
};
const defaultOptions = {
  disableExpectTypeSnapshotFix: false
};
const expect = createRule({
  create(context, [options]) {
    validate(context, options);
    return {};
  },
  defaultOptions: [defaultOptions],
  meta: {
    docs: {
      description: "Expects type error, type snapshot, or type.",
      requiresTypeChecking: true
    },
    fixable: "code",
    messages,
    schema: [
      {
        additionalProperties: false,
        properties: {
          disableExpectTypeSnapshotFix: {
            type: "boolean"
          }
        },
        type: "object"
      }
    ],
    type: "problem"
  },
  name: "expect"
});
function validate(context, options) {
  const parserServices = ESLintUtils.getParserServices(context);
  const { program } = parserServices;
  const fileName = context.getFilename();
  const sourceFile = program.getSourceFile(fileName);
  if (!sourceFile) {
    context.report({
      data: {
        fileName
      },
      loc: {
        column: 0,
        line: 1
      },
      messageId: "FileIsNotIncludedInTsconfig"
    });
    return;
  }
  if (!/\$Expect(?:Type|Error|\?)|\^\?/.test(sourceFile.text)) {
    return;
  }
  const checker = program.getTypeChecker();
  const diagnostics = ts.getPreEmitDiagnostics(program, sourceFile);
  const languageService = ts.createLanguageService(
    getLanguageServiceHost(program)
  );
  const {
    duplicates,
    errorLines,
    syntaxErrors,
    twoSlashAssertions,
    typeAssertions
  } = parseAssertions(sourceFile);
  for (const line of duplicates) {
    context.report({
      loc: {
        column: 0,
        line: line + 1
      },
      messageId: "Multiple$ExpectTypeAssertions"
    });
  }
  const seenDiagnosticsOnLine = new Set(
    diagnostics.filter(isDiagnosticWithStart).map((diagnostic) => lineOfPosition(diagnostic.start, sourceFile))
  );
  for (const line of errorLines) {
    if (!seenDiagnosticsOnLine.has(line)) {
      context.report({
        loc: {
          column: 0,
          line: line + 1
        },
        messageId: "ExpectedErrorNotFound"
      });
    }
  }
  for (const { line, type } of syntaxErrors) {
    context.report({
      data: {
        message: type === "MissingExpectType" ? '$ExpectType requires type argument (e.g. // $ExpectType "string")' : type === "MissingSnapshotName" ? "$ExpectTypeSnapshot requires snapshot name argument (e.g. // $ExpectTypeSnapshot MainComponentAPI)" : 'Invalid twoslash assertion; make sure there is a space after the "^?".'
      },
      loc: {
        column: 0,
        line: line + 1
      },
      messageId: "SyntaxError"
    });
  }
  for (const [, assertion] of typeAssertions) {
    if (assertion.assertionType === "snapshot") {
      assertion.expected = getTypeSnapshot(fileName, assertion.snapshotName);
    }
  }
  const { unmetExpectations, unusedAssertions } = getExpectTypeFailures(
    sourceFile,
    { twoSlashAssertions, typeAssertions },
    checker,
    languageService
  );
  for (const { actual, assertion, node } of unmetExpectations) {
    const templateDescriptor = {
      data: {
        actual,
        expected: assertion.expected
      },
      loc: loc(sourceFile, node)
    };
    if (assertion.assertionType === "snapshot") {
      const { snapshotName } = assertion;
      const start = node.getStart();
      context.report({
        ...templateDescriptor,
        // ESLint doesn't indicate whether it's in --fix mode, and fixers run immediately.
        // We don't have a way to delay file updates natively in ESLint.
        // Instead, we use this inaccurate heuristic to know if we're in CLI --fix...
        // See: https://github.com/JoshuaKGoldberg/eslint-plugin-expect-type/issues/14
        fix: process.argv.includes("--fix") ? () => {
          return {
            range: [start, start],
            get text() {
              if (!options.disableExpectTypeSnapshotFix) {
                updateTypeSnapshot(fileName, snapshotName, actual);
              }
              return "";
            }
          };
        } : void 0,
        messageId: typeof assertion.expected === "undefined" ? "TypeSnapshotNotFound" : "TypeSnapshotDoNotMatch"
      });
    } else {
      context.report({
        ...templateDescriptor,
        messageId: "TypesDoNotMatch",
        ...assertion.assertionType === "twoslash" ? {
          fix: () => {
            const { expectedPrefix, expectedRange, insertSpace } = assertion;
            return {
              range: expectedRange,
              text: (insertSpace ? " " : "") + actual.split("\n").map((line, i) => i > 0 ? expectedPrefix + line : line).join("\n")
            };
          }
        } : {}
      });
    }
  }
  for (const line of unusedAssertions) {
    context.report({
      loc: {
        column: 0,
        line: line + 1
      },
      messageId: "OrphanAssertion"
    });
  }
}
function parseAssertions(sourceFile) {
  const errorLines = /* @__PURE__ */ new Set();
  const typeAssertions = /* @__PURE__ */ new Map();
  const duplicates = [];
  const syntaxErrors = [];
  const twoSlashAssertions = [];
  const { text } = sourceFile;
  const commentRegexp = /\/\/(.*)/g;
  const lineStarts = sourceFile.getLineStarts();
  let curLine = 0;
  while (true) {
    const commentMatch = commentRegexp.exec(text);
    if (commentMatch === null) {
      break;
    }
    const comment = commentMatch[1];
    const matchExpect = /^ ?\$Expect(TypeSnapshot|Type|Error)( (.*))?$/.exec(
      comment
    );
    const commentIndex = commentMatch.index;
    const line = getLine(commentIndex);
    if (matchExpect) {
      const directive = matchExpect[1];
      const payload = matchExpect[3];
      switch (directive) {
        case "TypeSnapshot":
          const snapshotName = payload;
          if (snapshotName) {
            if (typeAssertions.delete(line)) {
              duplicates.push(line);
            } else {
              typeAssertions.set(line, {
                assertionType: "snapshot",
                snapshotName
              });
            }
          } else {
            syntaxErrors.push({
              line,
              type: "MissingSnapshotName"
            });
          }
          break;
        case "Error":
          if (errorLines.has(line)) {
            duplicates.push(line);
          }
          errorLines.add(line);
          break;
        case "Type": {
          const expected = payload;
          if (expected) {
            if (typeAssertions.delete(line)) {
              duplicates.push(line);
            } else {
              typeAssertions.set(line, { assertionType: "manual", expected });
            }
          } else {
            syntaxErrors.push({
              line,
              type: "MissingExpectType"
            });
          }
          break;
        }
      }
    } else {
      const assertion = parseTwoslashAssertion(
        comment,
        commentIndex,
        line,
        text,
        lineStarts
      );
      if (assertion) {
        if ("type" in assertion) {
          syntaxErrors.push(assertion);
        } else {
          twoSlashAssertions.push(assertion);
        }
      }
    }
  }
  return {
    duplicates,
    errorLines,
    syntaxErrors,
    twoSlashAssertions,
    typeAssertions
  };
  function getLine(pos) {
    while (lineStarts[curLine + 1] <= pos) {
      curLine++;
    }
    return isFirstOnLine(text, lineStarts[curLine], pos) ? curLine + 1 : curLine;
  }
}
function parseTwoslashAssertion(comment, commentIndex, commentLine, sourceText, lineStarts) {
  const matchTwoslash = /^( *)\^\?(.*)$/.exec(comment);
  if (!matchTwoslash) {
    return null;
  }
  const whitespace = matchTwoslash[1];
  const rawPayload = matchTwoslash[2];
  if (rawPayload.length && !rawPayload.startsWith(" ")) {
    return {
      line: commentLine - 1,
      type: "InvalidTwoslash"
    };
  }
  let expected = rawPayload.slice(1);
  if (commentLine === 1) {
    return {
      expected,
      expectedPrefix: "",
      expectedRange: [-1, -1],
      insertSpace: false,
      position: -1
    };
  }
  const caretIndex = commentIndex + whitespace.length + 2;
  const position = caretIndex - (lineStarts[commentLine - 1] - lineStarts[commentLine - 2]);
  const expectedRange = [
    commentIndex + whitespace.length + 5,
    commentLine < lineStarts.length ? lineStarts[commentLine] - 1 : sourceText.length
  ];
  const expectedPrefix = sourceText.slice(
    lineStarts[commentLine - 1],
    commentIndex + 2 + whitespace.length
  ) + "   ";
  for (let nextLine = commentLine; nextLine < lineStarts.length; nextLine++) {
    const thisLineEnd = nextLine + 1 < lineStarts.length ? lineStarts[nextLine + 1] - 1 : sourceText.length;
    const lineText = sourceText.slice(lineStarts[nextLine], thisLineEnd + 1);
    if (lineText.startsWith(expectedPrefix)) {
      if (nextLine === commentLine) {
        expected += "\n";
      }
      expected += lineText.slice(expectedPrefix.length);
      expectedRange[1] = thisLineEnd;
    } else {
      break;
    }
  }
  let insertSpace = false;
  if (expectedRange[0] > expectedRange[1]) {
    expectedRange[0] = expectedRange[1];
    insertSpace = true;
  }
  return { expected, expectedPrefix, expectedRange, insertSpace, position };
}
function isFirstOnLine(text, lineStart, pos) {
  for (let i = lineStart; i < pos; i++) {
    if (/\S/.test(text[i])) {
      return false;
    }
  }
  return true;
}
function matchReadonlyArray(actual, expected) {
  if (!(/\breadonly\b/.test(actual) && /\bReadonlyArray\b/.test(expected))) {
    return false;
  }
  const readonlyArrayRegExp = /\bReadonlyArray</y;
  const readonlyModifierRegExp = /\breadonly /y;
  let expectedPos = 0;
  let actualPos = 0;
  let depth = 0;
  while (expectedPos < expected.length && actualPos < actual.length) {
    const expectedChar = expected.charAt(expectedPos);
    const actualChar = actual.charAt(actualPos);
    if (expectedChar === actualChar) {
      expectedPos++;
      actualPos++;
      continue;
    }
    if (depth > 0 && expectedChar === ">" && actualChar === "[" && actualPos < actual.length - 1 && actual.charAt(actualPos + 1) === "]") {
      depth--;
      expectedPos++;
      actualPos += 2;
      continue;
    }
    readonlyArrayRegExp.lastIndex = expectedPos;
    readonlyModifierRegExp.lastIndex = actualPos;
    if (readonlyArrayRegExp.test(expected) && readonlyModifierRegExp.test(actual)) {
      depth++;
      expectedPos += 14;
      actualPos += 9;
      continue;
    }
    return false;
  }
  return true;
}
function getLanguageServiceHost(program) {
  return {
    getCompilationSettings: () => program.getCompilerOptions(),
    getCurrentDirectory: () => program.getCurrentDirectory(),
    getDefaultLibFileName: (options) => ts.getDefaultLibFilePath(options),
    getScriptFileNames: () => program.getSourceFiles().map((sourceFile) => sourceFile.fileName),
    getScriptSnapshot: (name) => ts.ScriptSnapshot.fromString(program.getSourceFile(name)?.text ?? ""),
    getScriptVersion: () => "1",
    // NB: We can't check `program` for files, it won't contain valid files like package.json
    /* eslint-disable @typescript-eslint/unbound-method */
    directoryExists: ts.sys.directoryExists,
    fileExists: ts.sys.fileExists,
    getDirectories: ts.sys.getDirectories,
    readDirectory: ts.sys.readDirectory,
    readFile: ts.sys.readFile
    /* eslint-enable @typescript-eslint/unbound-method */
  };
}
function getExpectTypeFailures(sourceFile, assertions, checker, languageService) {
  const { twoSlashAssertions, typeAssertions } = assertions;
  const unmetExpectations = [];
  ts.forEachChild(sourceFile, function iterate(node) {
    const line = lineOfPosition(node.getStart(sourceFile), sourceFile);
    const assertion = typeAssertions.get(line);
    if (assertion !== void 0) {
      const { expected } = assertion;
      let nodeToCheck = node;
      if (node.kind === ts.SyntaxKind.ExpressionStatement) {
        node = node.expression;
      }
      nodeToCheck = getNodeForExpectType(node);
      const type = checker.getTypeAtLocation(nodeToCheck);
      const actual = checker.typeToString(
        type,
        /*enclosingDeclaration*/
        void 0,
        ts.TypeFormatFlags.NoTruncation
      );
      if (!expected || actual !== expected && !matchReadonlyArray(actual, expected)) {
        unmetExpectations.push({ actual, assertion, node });
      }
      typeAssertions.delete(line);
    }
    ts.forEachChild(node, iterate);
  });
  const twoSlashFailureLines = [];
  if (twoSlashAssertions.length) {
    for (const assertion of twoSlashAssertions) {
      const { expected, position } = assertion;
      if (position === -1) {
        twoSlashFailureLines.push(0);
        continue;
      }
      const node = getNodeAtPosition(sourceFile, position);
      if (!node) {
        twoSlashFailureLines.push(
          sourceFile.getLineAndCharacterOfPosition(position).line
        );
        continue;
      }
      const qi = languageService.getQuickInfoAtPosition(
        sourceFile.fileName,
        node.getStart()
      );
      if (!qi?.displayParts) {
        twoSlashFailureLines.push(
          sourceFile.getLineAndCharacterOfPosition(position).line
        );
        continue;
      }
      const actual = qi.displayParts.map((dp) => dp.text).join("");
      if (!matchModuloWhitespace(actual, expected)) {
        unmetExpectations.push({
          actual,
          assertion: { assertionType: "twoslash", ...assertion },
          node
        });
      }
    }
  }
  return {
    unmetExpectations,
    unusedAssertions: [...twoSlashFailureLines, ...typeAssertions.keys()]
  };
}
function getNodeAtPosition(sourceFile, position) {
  let candidate = void 0;
  ts.forEachChild(sourceFile, function iterate(node) {
    const start = node.getStart();
    const end = node.getEnd();
    if (position >= start && position <= end) {
      candidate = node;
      ts.forEachChild(node, iterate);
    }
  });
  return candidate;
}
function matchModuloWhitespace(actual, expected) {
  const normActual = actual.replace(/[\n\r ]+/g, " ").trim();
  const normExpected = expected.replace(/[\n\r ]+/g, " ").trim();
  return normActual === normExpected;
}
function getNodeForExpectType(node) {
  if (ts.isVariableStatement(node)) {
    const {
      declarationList: { declarations }
    } = node;
    if (declarations.length === 1) {
      const { initializer } = declarations[0];
      if (initializer) {
        return initializer;
      }
    }
  }
  return node;
}
function lineOfPosition(pos, sourceFile) {
  return sourceFile.getLineAndCharacterOfPosition(pos).line;
}
export {
  expect
};
//# sourceMappingURL=expect.mjs.map